import { ZipReader, fromDataSource } from '../utils/index.js';
import { CryptoService } from '../crypto/declarations.js';
import { type AuthProvider, AppIdAuthProvider, HttpRequest, withHeaders } from '../../../src/auth/auth.js';
import EAS from '../../../src/auth/Eas.js';
import { EncryptParams, DecryptParams } from './builders.js';
import { DecoratedReadableStream } from './DecoratedReadableStream.js';
import { DecryptParamsBuilder, type DecryptSource, EncryptParamsBuilder } from './builders.js';
import { KasPublicKeyInfo, OriginAllowList } from '../../../src/access.js';
export declare const uploadBinaryToS3: (stream: ReadableStream<Uint8Array>, uploadUrl: string, fileSize: number) => Promise<void>;
export interface ClientConfig {
    cryptoService?: CryptoService;
    organizationName?: string;
    clientId?: string;
    dpopEnabled?: boolean;
    dpopKeys?: Promise<CryptoKeyPair>;
    kasEndpoint?: string;
    /**
     * Service to use to look up ABAC. Used during autoconfigure. Defaults to
     * kasEndpoint without the trailing `/kas` path segment, if present.
     */
    policyEndpoint?: string;
    /**
     * List of allowed KASes to connect to for rewrap requests.
     * Defaults to `[kasEndpoint]`.
     */
    allowedKases?: string[];
    ignoreAllowList?: boolean;
    easEndpoint?: string;
    keyRewrapEndpoint?: string;
    keyUpsertEndpoint?: string;
    refreshToken?: string;
    kasPublicKey?: string;
    oidcOrigin?: string;
    externalJwt?: string;
    authProvider?: AuthProvider | AppIdAuthProvider;
    readerUrl?: string;
    entityObjectEndpoint?: string;
    fileStreamServiceWorker?: string;
    progressHandler?: (bytesProcessed: number) => void;
}
export declare function createSessionKeys({ authProvider, cryptoService, dpopKeys, }: {
    authProvider?: AuthProvider | AppIdAuthProvider;
    cryptoService: CryptoService;
    dpopKeys?: Promise<CryptoKeyPair>;
}): Promise<CryptoKeyPair>;
export declare class Client {
    readonly cryptoService: CryptoService;
    /**
     * Default kas endpoint, if present. Required for encrypt.
     */
    readonly kasEndpoint: string;
    /**
     * Policy service endpoint, if present.
     * Required for autoconfiguration with ABAC.
     */
    readonly policyEndpoint: string;
    /**
     * List of allowed KASes to connect to for rewrap requests.
     * Defaults to `[this.kasEndpoint]`.
     */
    readonly allowedKases: OriginAllowList;
    readonly kasKeys: Record<string, Promise<KasPublicKeyInfo>>;
    readonly easEndpoint?: string;
    readonly clientId?: string;
    readonly authProvider?: AuthProvider | AppIdAuthProvider;
    readonly readerUrl?: string;
    readonly fileStreamServiceWorker?: string;
    /**
     * Session binding keys. Used for DPoP and signed request bodies.
     */
    readonly dpopKeys: Promise<CryptoKeyPair>;
    readonly eas?: EAS;
    readonly dpopEnabled: boolean;
    readonly clientConfig: ClientConfig;
    /**
     * An abstraction for protecting and accessing data using TDF3 services.
     * @param {Object} [config.keypair] - keypair generated for signing. Optional, will be generated by sdk if not passed
     * @param {String} [config.clientId]
     * @param {String} [config.kasEndpoint] - Key Access Server url
     * @param {String} [config.refreshToken] - After logging in to browser OIDC interface user
     * receives fresh token that needed by SDK for auth needs
     * @param {String} [config.externalJwt] - JWT from external authority (eg Google)
     * @param {String} [config.oidcOrigin] - Endpoint of authentication service
     */
    constructor(config: ClientConfig);
    /**
     * Encrypt plaintext into TDF ciphertext. One of the core operations of the Virtru SDK.
     *
     * @param scope dissem and attributes for constructing the policy
     * @param source source object of unencrypted data
     * @param [asHtml] If we should wrap the TDF data in a self-opening HTML wrapper. Defaults to false
     * @param [autoconfigure] If we should use scope.attributes to configure KAOs
     * @param [metadata] Additional non-secret data to store with the TDF
     * @param [opts] Test only
     * @param [mimeType] mime type of source. defaults to `unknown`
     * @param [offline] Where to store the policy. Defaults to `false` - which results in `upsert` events to store/update a policy
     * @param [windowSize] - segment size in bytes. Defaults to a a million bytes.
     * @param [keyMiddleware] - function that handle keys
     * @param [streamMiddleware] - function that handle stream
     * @param [eo] - (deprecated) entity object
     * @return a {@link https://nodejs.org/api/stream.html#stream_class_stream_readable|Readable} a new stream containing the TDF ciphertext
     */
    encrypt({ scope, autoconfigure, source, asHtml, metadata, mimeType, offline, windowSize, eo, keyMiddleware, streamMiddleware, splitPlan, assertionConfigs, }: EncryptParams): Promise<DecoratedReadableStream>;
    /**
     * Decrypt TDF ciphertext into plaintext. One of the core operations of the Virtru SDK.
     *
     * @param params keyMiddleware fucntion to process key
     * @param params streamMiddleware fucntion to process streamMiddleware
     * @param params.source A data stream object, one of remote, stream, buffer, etc. types.
     * @param params.eo Optional entity object (legacy AuthZ)
     * @param params.assertionVerificationKeys Optional verification keys for assertions.
     * @return a {@link https://nodejs.org/api/stream.html#stream_class_stream_readable|Readable} stream containing the decrypted plaintext.
     * @see DecryptParamsBuilder
     */
    decrypt({ eo, source, keyMiddleware, streamMiddleware, assertionVerificationKeys, noVerifyAssertions, concurrencyLimit, }: DecryptParams): Promise<DecoratedReadableStream>;
    /**
     * Get the unique policyId associated with TDF ciphertext. Useful for managing authorization policies of encrypted data.
     * <br/><br/>
     * The policyId is embedded in the ciphertext so this is a local operation.
     *
     * @param {object} source - Required. TDF data stream,
     * generated using {@link DecryptParamsBuilder#build|DecryptParamsBuilder's build()}.
     * @return {string} - the unique policyId, which can be used for tracking purposes or policy management operations.
     * @see DecryptParamsBuilder
     */
    getPolicyId({ source }: {
        source: DecryptSource;
    }): Promise<any>;
    loadTDFStream({ source }: {
        source: DecryptSource;
    }): Promise<{
        manifest: import("../models/manifest.js").Manifest;
        zipReader: ZipReader;
        centralDirectory: import("../utils/zip-reader.js").CentralDirectory[];
    }>;
}
export type { AuthProvider };
export { AppIdAuthProvider, DecryptParamsBuilder, DecryptSource, EncryptParamsBuilder, HttpRequest, fromDataSource, withHeaders, };
//# sourceMappingURL=index.d.ts.map